#include <u.h>
#include <libc.h>
#include <fcall.h>
#include <thread.h>
#include <libsec.h>
#include <pool.h>
#include <9p.h>
#include <bio.h>
#include <xml.h>
#include "wdfs.h"

enum {
	Maxents = 8192,		/* allow this many entries without reaping */
	Percent = 5,		/* reap entries older than Percent of time since last file change */
	Mintime = 60,		/* min secs we will cache file for */
	Maxtime = 60*60		/* max secs we will cache file for */
};

static Cache *
fillcache(Session *s, char *path, int dirscan, long now)
{
	Cache *c;
	long newest;
	int i, ndir;
	Dir *dir, *d;

	dir = nil;
	ndir = _wd_dirread(s, path, dirscan, &dir);
	c = emalloc9p(sizeof(Cache));
	c->path = estrdup9p(path);
	c->dirscan = dirscan;
	c->dir = dir;
	c->ndir = ndir;
	c->atime = now;

	newest = 0x7fffffff;			/* assuming 32 bit ints */
	for(d = dir, i = 0; i < ndir; i++, d++)
		if(newest > now - d->mtime)
			newest = now - d->mtime;
	if(newest < Mintime)
		newest = Mintime;
	if(newest > Maxtime)
		newest = Maxtime;
	if(ndir == 0)
		newest = Mintime;		/* don't cache negitive results for long */
	c->expiry = now + (newest * Percent) / 100;
	return c;
}

static Dir *
cpdir(Dir *dir, int ndir)
{
	int i;
	Dir *new, *d, *n;

	if(ndir < 0)
		return nil;

	new = emalloc9p(sizeof(Dir) * ndir);
	for(d = dir, n = new, i = 0; i < ndir; i++, d++, n++){
		*n = *d;
		n->name = estrdup9p(d->name);
		n->muid = estrdup9p(d->muid);
		n->uid = estrdup9p(d->uid);
		n->gid = estrdup9p(d->gid);
	}
	return new;
}


int
wd_dirread(Session *s, char *path, int dirscan, Dir **dp)
{
	long now;
	Cache *c;

	now = time(nil);
	rlock(&s->lock);
	for(c = s->cache; c; c = c->next)
		if(strcmp(c->path, path) == 0 && c->dirscan == dirscan)
			break;
	if(c == nil){
		if(Wddebug)
			fprint(2, "cache: path=%s dirscan=%d - miss\n", path, dirscan);

		runlock(&s->lock);
		if((c = fillcache(s, path, dirscan, now)) == nil)
			return -1;

		wlock(&s->lock);
		c->next = s->cache;
		s->cache = c;
		if(c->ndir < 0)
			s->ncache += 1;
		else
			s->ncache += c->ndir;
		wunlock(&s->lock);
		rlock(&s->lock);
	}
	else
	if(Wddebug)
		fprint(2, "cache: path=%s dirscan=%d - hit ndir=%d\n", path, dirscan, c->ndir);

	*dp = cpdir(c->dir, c->ndir);
	c->atime = now;
	runlock(&s->lock);
	return c->ndir;
}

static Cache *
release(Session *s, Cache **p, Cache *c)
{
	*p = c->next;
	if(c->ndir < 0)
		s->ncache -= 1;
	else
		s->ncache -= c->ndir;
	freedir(c->dir, c->ndir);
	free(c->path);
	free(c);
	return *p;
}

void
invalidate(Session *s, char *path)
{
	int len;
	char *slash;
	Cache **p, *c;

	assert((slash = strrchr(path, '/')) != nil);
	len = slash-path;
	if(len > 1)		/* not root dir */
		len--;
	
	wlock(&s->lock);
	for(p = &s->cache, c = s->cache; c; p = &c->next, c = c->next){
		if(strcmp(c->path, path) == 0 && !c->dirscan)
			if((c = release(s, p, c)) == nil)
				break;
		if(strncmp(c->path, path, len) == 0 && c->dirscan)
			if((c = release(s, p, c)) == nil)
				break;
	}
	wunlock(&s->lock);

}

void
flushcache(Session *s)
{
	long now, sec;
	Cache *c, **p;

	while(1){
		sleep(10000);
		now = time(nil);

		/*
		 * Entries are considered stale if they are older than 
		 * "Percent" (5% by default) of the age of the object.
		 * for more info see:
		 *
		 * %A Vincent Cate.
		 * %T Alex - a Global Filesystem.
		 * %J In Proceedings of the 1992 USENIX File System Workshop
		 * %P 1-12
		 * %D May 1992
		 */
		wlock(&s->lock);
		for(p = &s->cache, c = s->cache; c; p = &c->next, c = c->next)
			if(now > c->expiry)
				if((c = release(s, p, c)) == nil)
					break;

		wunlock(&s->lock);

		/*
		 * Start stealing back entries when the number of cached Dirs
		 * is greater than "Maxents". We start with objects accessed more
		 * than "Maxtime" secconds ago.
		 *
		 * We keep retrying, dividing the timeout by two until the number
		 * of entries in the cache gets below "Maxents"
		 *
		 * This is to prevent us from using too much RAM.
		 *
		 * NB: s->ncache counts the number of Dir structures used not the
		 * number of paths cached, thus it better expresses RAM use.
		 */
		sec = Maxtime;
		while(s->ncache > Maxents){
			wlock(&s->lock);
			for(p = &s->cache, c = s->cache; c; p = &c->next, c = c->next)
				if((now - c->atime) > sec)
					if((c = release(s, p, c)) == nil)
						break;

			wunlock(&s->lock);
			sec /= 2;	
		}
	}
}
