#include <u.h>
#include <libc.h>
#include <fcall.h>
#include <thread.h>
#include <libsec.h>
#include <pool.h>
#include <9p.h>
#include <bio.h>
#include <xml.h>
#include "wdfs.h"

int Wddebug;
int Propdebug;
int Partget;
int Partput;

static Session *Sess;		/* current session */
static int Flusher;		/* PID of cache flush thread */

static void
responderrstr(Req *r)
{
	char e[ERRMAX];
	*e = 0;
	rerrstr(e, sizeof e);
	respond(r, e);
}

static char *
newpath(char *path, char *name)
{
	char *p, *q, *npath;

	assert((p = strrchr(path, '/')) != nil);

	if(strcmp(name, "..") == 0){
		if(p == path)
			return estrdup9p("/");
		q = emalloc9p((p-path)+1);
		strecpy(q, q+(p-path)+1, path);
		return q;
	}

	if(strcmp(path, "/") == 0)
		npath = smprint("/%s", name);
	else
		npath = smprint("%s/%s", path, name);
	assert(npath != nil);
	return npath;
}

void
freedir(Dir *d, int n)
{
	int i;

	for(i = 0; i < n; i++){
		free(d[i].name);
		free(d[i].muid);
		free(d[i].uid);
		free(d[i].gid);
	}
	free(d);
}


static int
dirgen(int slot, Dir *d, void *aux)
{
	Aux *a;

	a = aux;
	if(a->ndir <= 0){
		/*
		 * re-read as the dir info read at open
		 * must have expired from the cache
		 */
		if((a->ndir = wd_dirread(Sess, a->path, 1, &a->dir)) < 0)
			return -1;
	}
	if(slot >= a->ndir)
		return -1;
	*d = a->dir[slot];
	d->name = estrdup9p(d->name);
	d->muid = estrdup9p(d->muid);
	d->uid = estrdup9p(d->uid);
	d->gid = estrdup9p(d->gid);
	return 0;
}

static void
fsattach(Req *r)
{
	int n;
	Aux *a;
	Dir *d; 
	char *spec, *root;

	root = "/";

	spec = r->ifcall.aname;
	if(spec && *spec){
		respond(r, "unwanted attach specifier");
		return;
	}

	if((n = wd_dirread(Sess, root, 0, &d)) < 0){
		responderrstr(r);
		return;
	}

	a = emalloc9p(sizeof(Aux));
	memset(a, 0, sizeof(Aux));
	a->path = estrdup9p(root);

	r->ofcall.qid = d->qid;
	r->fid->qid = d->qid;
	r->fid->aux = a;
	freedir(d, n);
	respond(r, nil);
}

static char*
fsclone(Fid *ofid, Fid *fid)
{
	Aux *oa, *a;

	a = emalloc9p(sizeof(Aux));
	memset(a, 0, sizeof(Aux));

	oa = ofid->aux;
	a->path = estrdup9p(oa->path);
	fid->aux = a;
	return nil;
}

static char*
fswalk1(Fid *fid, char *name, Qid *qidp)
{
	int n;
	Dir *d;
	Aux *a;
	char *npath;


	a = fid->aux;
	npath = newpath(a->path, name);
	if((n = wd_dirread(Sess, npath, 0, &d)) < 0){
		free(npath);
		return "not found";
	}

	free(a->path);	
	a->path = npath;
	*qidp = d->qid;
	freedir(d, n);
	return nil;
}

static void
fsstat(Req *r)
{
	Aux *a;
	Dir *d;

	a = r->fid->aux;
	if(wd_dirread(Sess, a->path, 0, &d) != 1){
		responderrstr(r);
		return;
	}
	r->d = *d;
	free(d);
	respond(r, nil);
	
}

static void
fscreate(Req *r)
{
	int n;
	Dir *d;
	char *npath;
	Aux *a = r->fid->aux;

	npath = newpath(a->path, r->ifcall.name);
	if((r->ifcall.perm & DMDIR) == DMDIR){
		if(wd_mkcol(Sess, npath) == -1){
			free(npath);
			responderrstr(r);
			return;
		}
	}
	else{
		if((a->wdf = wd_create(Sess, npath, r->ifcall.mode)) == nil){
			free(npath);
			responderrstr(r);
			return;
		}
	}

	if(wd_wstat(Sess, npath, nil, ~0LL, ~0L, ~0L, r->ifcall.perm) < 0){
		free(npath);
		responderrstr(r);
		return;
	}

	/*
	 * sadly we need to do a wd_dirread to get the server's
	 * view of the last modified time of the file for qid.vers
	 */
	if((n = wd_dirread(Sess, npath, 0, &d)) < 0){
		free(npath);
		responderrstr(r);
		return;
	}

	free(a->path);
	a->path = npath;
	invalidate(Sess, a->path);
	r->fid->qid = d->qid;
	r->ofcall.qid = d->qid;
	freedir(d, n);
	respond(r, nil);
}

static void
fsopen(Req *r)
{
	Aux *a;

	a = r->fid->aux;
	if((r->fid->qid.type & QTDIR) == QTDIR){
		if((a->ndir = wd_dirread(Sess, a->path, 1, &a->dir)) < 0)
			responderrstr(r);
		else
			respond(r, nil);
		return;
	}

	if((a->wdf = wd_open(Sess, a->path, r->ifcall.mode)) == nil){
		responderrstr(r);
		return;
	}

	respond(r, nil);
	return;

}

static void
fsread(Req *r)
{
	int n;
	Aux *a;

	a = r->fid->aux;
	if(r->fid->qid.type & QTDIR){
		dirread9p(r, dirgen, a);
		respond(r, nil);
		return;
	}

	if((n = wd_pread(a->wdf, r->ofcall.data,
		r->ifcall.count, r->ifcall.offset)) < 0){
		responderrstr(r);
		return;
	}
	r->ofcall.count	= n;
	respond(r, nil);
}

static void
fswrite(Req *r)
{
	int n;
	Aux *a;

	a = r->fid->aux;
	if((n = wd_pwrite(a->wdf, r->ifcall.data,
			r->ifcall.count, r->ifcall.offset)) < 0){
		responderrstr(r);
		return;
	}
	r->ofcall.count	= n;
	respond(r, nil);
}

static void
fsdestroyfid(Fid *f)
{
	Aux *a;

	a = f->aux;
	if(a == nil)
		return;
	if(a->wdf)
		wd_close(a->wdf);
	freedir(a->dir, a->ndir);
	free(a->path);
	free(a);
}

static void
fsremove(Req *r)
{
	Aux *a;

	a = r->fid->aux;
	if(wd_delete(Sess, a->path) < 0){
		responderrstr(r);
		return;
	}
	invalidate(Sess, a->path);
	respond(r, nil);
}

static void
fswstat(Req *r)
{
	Aux *a;
	char *p, *npath;

	a = r->fid->aux;
	invalidate(Sess, a->path);
	if(r->d.name && r->d.name[0]){
		if((p = strrchr(a->path, '/')) == nil){
			respond(r, "bad path");
			return;
		}
		npath = emalloc9p((p-a->path)+strlen(r->d.name)+2);
		strecpy(npath, npath+(p- a->path)+2, a->path);
		strcat(npath, r->d.name);

		if(wd_move(Sess, a->path, npath) == -1){
			werrstr("(rename) - %r");
			responderrstr(r);
			free(npath);
			return;
		}
		free(a->path);
		a->path = npath;
		invalidate(Sess, a->path);
	}

	if(r->d.gid && r->d.gid[0]){	/* FIXME: maybe this is how we can do a checkin/checkout for SNV repositories?? */
		respond(r, "cannot change version-name");
		return;
	}

	if(wd_wstat(Sess, a->path, r->d.uid, r->d.length,
	    r->d.mtime, r->d.atime, r->d.mode) < 0){
		responderrstr(r);
		return;
	}

	respond(r, nil);
}

static void
fsend(Srv *srv)
{
	USED(srv);
	wd_hangup(Sess);
	postnote(PNPROC, Flusher, "die");
}

Srv fs = {
	.destroyfid =	fsdestroyfid,
	.attach=	fsattach,
	.open=		fsopen,
	.create=	fscreate,
	.read=		fsread,
	.write=		fswrite,
	.remove=	fsremove,
	.stat=		fsstat,
	.wstat=		fswstat,
	.clone= 	fsclone,
	.walk1= 	fswalk1,
	.end=		fsend
};

static void
ding(void *u, char *msg)
{
	USED(u);

	if(strstr(msg, "alarm"))
		noted(NCONT);
	noted(NDFLT);
}

static void
usage(void)
{
	fprint(2, "usage: %s [-D names] [-w webfs] [-m mntpt] [-s srvname] [-p ge] [-p put] url\n", argv0);
	exits("usage");
}

void
main(int argc, char **argv)
{
	char *mtpt, *svs, *webfs, *p, *s;

	svs = nil;
	mtpt = nil;
	webfs = "/mnt/web";
	ARGBEGIN{
	case 'D':
		s = EARGF(usage());
		if(strstr(s, "prop"))
			Propdebug++;
		else
		if(strstr(s, "mem"))
			mainmem->flags |= POOL_NOREUSE|POOL_ANTAGONISM|POOL_PARANOIA;
		else
		if(strstr(s, "xml"))
			xml_debug++;
		else
		if(strstr(s, "wd"))
			Wddebug++;
		else
		if(strstr(s, "9p"))
			chatty9p++;
		else
			sysfatal("unknown debug keyword, need one of: prop mem xml wd 9p\n");
		break;
	case 'p':
		s = EARGF(usage());
		if(strstr(s, "get"))
			Partget++;
		if(strstr(s, "put"))
			Partput++;
		break;
	case 'm':
		mtpt = EARGF(usage());
		break;
	case 's':
		svs = EARGF(usage());
		break;
	case 'w':
		webfs = EARGF(usage());
		break;
	default:
		fprint(2, "Unrecognized option.\n");
		usage();
	}ARGEND

	if(argc != 1)
		usage();

	notify(ding);
	
	fmtinstall('D', dirfmt);
	fmtinstall('D', dirmodefmt);
	fmtinstall('F', fcallfmt);

	if(!mtpt && !svs)
		mtpt = "/n/wdfs";

	if((p = strrchr(argv[0], '/')) != nil && strcmp(p, "/") == 0)
		*p = 0;

	if(strncmp(argv[0], "http://", 7) != 0 &&
	   strncmp(argv[0], "https://", 8) != 0)
		sysfatal("%s - not a valid url\n", argv[0]);

	if((Sess = wd_dial(webfs, argv[0])) == nil)
		sysfatal("%s cannot connect - %r\n", argv[0]);

	Flusher = rfork(RFPROC|RFMEM|RFNOTEG|RFNAMEG);
	switch(Flusher){
	case -1:
		sysfatal("cannot fork %r\n");
	case 0:
		flushcache(Sess);
		break;
	default:
		break;
	}

	postmountsrv(&fs, svs, mtpt, MREPL|MCREATE);
	exits(nil);
}
