#include <u.h>
#include <libc.h>
#include <fcall.h>
#include <thread.h>
#include <9p.h>
#include <libsec.h>
#include <ctype.h>
#include <flate.h>
#include <bio.h>
#include <xml.h>
#include "wdfs.h"

static int
testbool(char *s)
{
	if(cistrcmp(s, "true") == 0)
		return 1;
	if(cistrcmp(s, "t") == 0)
		return 1;
	if(strtoul(s, nil, 0) == 1)
		return 1;
	if(cistrcmp(s, "false") == 0)
		return 0;
	if(cistrcmp(s, "f") == 0)
		return 0;
	if(strtoul(s, nil, 0) == 0)
		return 0;
	return -1;
}

static int
status(Elem *base)
{
	Elem *ep;
	char *a[8];

	if(base == nil)
		return -1;
	for(ep = base->child; ep; ep = ep->next)
		if(strcmp(ep->name, "status") == 0 && ep->pcdata)
			if(tokenize(ep->pcdata, a, nelem(a)) >= 2)
				return atoi(a[1]);
	return -1;
}
			
static void
tmpfile(Wdfile *wdf)
{
	int i, fd;
	char buf[128];
	uint rnd[sizeof(int)*16];

	if(wdf->partialput && wdf->partialget)
		return;
	if(wdf->partialput && (wdf->mode & OWRITE|OREAD) == OWRITE)
		return;
	if(wdf->partialget && (wdf->mode & OWRITE|OREAD) == OREAD)
		return;

	fd = -1;
	do{
		genrandom((uchar *)rnd, sizeof(rnd));
		for(i = 0; i < nelem(rnd); i++){
			snprint(buf, sizeof(buf), "/tmp/wdfs.%ud", rnd[i]);
			if((fd = create(buf, ORDWR|ORCLOSE, 0)) != -1)
				break;
		}
	}while(fd == -1);
	wdf->tmpfd = fd;
}

static Elem *
decend(Xml *xp, Elem *base, char *name)
{
	Elem *ep;

	if((ep = xml_find(xp, base, name)) == nil)
		return nil;
	return ep->child;
}

static char *
pcdata(Xml *xp, Elem *base, char *path)
{
	Elem *ep;

	if((ep = xml_find(xp, base, path)) == nil)
		return nil;
	return ep->pcdata;
}

static Qid
mkqid(char *name, long mtime, int isdir)
{
	Qid q;
	uchar digest[SHA1dlen];

	sha1((uchar *)name, strlen(name), digest, nil);
	q.type = (isdir)? QTDIR: 0;
	q.vers = mtime;
	q.path = *((uvlong *)digest) & ~0xfL;
	return q;
}

static int
dhex(char c)
{
	if('0' <= c && c <= '9')
		return c-'0';
	if('a' <= c && c <= 'f')
		return c-'a'+10;
	if('A' <= c && c <= 'F')
		return c-'A'+10;
	return 0;
}

static char *
href2name(char *name)
{
	char *p, *w;
	uchar x;

	/* remove fragment */
	if((p = strchr(name, '#')) != nil)
		*p = 0;
	/* strip trailing slash */
	if((p = strrchr(name, '/')) != nil && strcmp(p, "/") == 0)
		*p = 0;
	/* just the trailing basename */
	if((p = strrchr(name, '/')) != nil)
		name = p+1;
	/* unescape name */
	for(p = w = name; x = *p; p++){
		if(x=='%' && isxdigit(p[1]) && isxdigit(p[2])){
			x = (dhex(p[1])<<4)|dhex(p[2]);
			if(x == 0 || (x > 0x1F && x < 0x7F && x == '/'))
				x = *p;
			else
				p += 2;
		} else if(x=='+')
			x = ' ';
		*w++ = x;
	}
	*w = 0;

	return name;
}

Session *
wd_dial(char *webfs, char *url)
{
	int fd, n;
	Session *s;
	char buf[128];
	Param params[] = {
		{ "url", 	"" },
		{ "request",	"GET" },
//		{ "request",	"OPTIONS" },
		{ nil, nil },
	};

	s = emalloc9p(sizeof(Session));

	snprint(buf, sizeof buf, "%s/clone", webfs);
	if((s->fd = open(buf, ORDWR)) < 0){
		free(s);
		return nil;
	}

	if((n = read(s->fd, buf, sizeof buf-1)) < 0){
		close(s->fd);
		free(s);
		return nil;
	}
	if(n == 0){
		werrstr("short read of clone file");
		close(s->fd);
		free(s);
		return nil;
	}
	buf[n] = '\0';
	s->conid = atoi(buf);
	s->base = estrdup9p(url);
	s->webfs = estrdup9p(webfs);

	if((fd = exch(s, params, nil, 0, -1)) < 0){
		close(s->fd);
		free(s->base);
		free(s);
		return nil;
	}
	close(fd);
	return s;
}

int
wd_hangup(Session *s)
{
	close(s->fd);
	free(s->base);
	free(s);
	return 0;
}

Wdfile *
wd_create(Session *s, char *path, int mode)
{
	int fd;
	Wdfile *wdf;
	Param params[] = {
		{ "url", 	path },
		{ "request",	"PUT" },
		{ "content",	"application/octet-stream" },
		{ "headers",	"Content-Length: 0\r\nContent-Range: 0-0/1" },
		{ nil, nil },
	};

	if(Wddebug)
		fprint(2, "wd_create: path=%s\n", path);

	invalidate(s, path);

	/* create a zero byte placeholder */
	if((fd = exch(s, params, nil, 0, -1)) < 0)
		return nil;
	close(fd);

	wdf = emalloc9p(sizeof(Wdfile));
	wdf->path = estrdup9p(path);
	wdf->partialput = Partput;
	wdf->partialget = Partget;
	wdf->mode = mode;
	wdf->sess = s;
	tmpfile(wdf);

	if(Wddebug)
		fprint(2, "	filecached=%s\n", (wdf->tmpfd)? "yes": "no");

	return wdf;
}

Wdfile *
wd_open(Session *s, char *path, int mode)
{
	Wdfile *wdf;
	int n, fd, got, len;
	char request[32], buf[8192];
	Param params[] = {
		{ "url", 		path },
		{ "request",	request },
		{ "headers",	"Translate: f" },		/* for MS exchange */
		{ nil, nil },
	};


	if(Wddebug)
		fprint(2, "wd_open: path=%s\n", path);

	if(Partget)
		strcpy(request, "HEAD");
	else
		strcpy(request, "GET");

	if((fd = exch(s, params, nil, -1, -1)) < 0)
		return nil;

	wdf = emalloc9p(sizeof(Wdfile));
	wdf->path = estrdup9p(path);
	if((len = contentlength(s)) != -1)
		wdf->length = len;
	wdf->partialput = Partput;
	wdf->partialget = Partget;
	wdf->mode = mode;
	wdf->sess = s;
	tmpfile(wdf);

	if((mode & OTRUNC) == 0 && wdf->tmpfd != 0){
		got = 0;
		while((n = read(fd, buf, sizeof(buf))) > 0){
			got += n;
			if(write(wdf->tmpfd, buf, n) < 0){
				close(wdf->tmpfd);
				free(wdf->path);
				free(wdf);
				close(fd);
				return nil;
			}
		}

		if(got != wdf->length){
			if(wdf->length != 0)	/* liars */
				fprint(2, "wd_open: read-ahead: content length wrong: %lld != %d\n",
					wdf->length, got);
			wdf->length = got;
		}
	}

	if(Wddebug)
		fprint(2, "	filecached=%s\n", (wdf->tmpfd)? "yes": "no");

	close(fd);
	return wdf;
}

int
wd_pread(Wdfile *wdf, void *buf, long len, uvlong off)
{
	int fd, n, got;
	char *p, range[128];
	Param params[] = {
		{ "url", 			wdf->path },
		{ "headers",		"Translate: f" },		/* for MS exchange */
		{ "request",		"GET" },
		{ "headers",		range },
		{ nil, nil },
	};

	if(Wddebug){
		fprint(2, "wr_pread: path=%s len=%ld off=%lld\n", wdf->path, len, off);
		fprint(2, "	filecached=%s\n", (wdf->tmpfd)? "yes": "no");
	}

	if(off >= wdf->length)		/* EOF */
		return 0;

	if(wdf->tmpfd)
		return pread(wdf->tmpfd, buf, len, off);

	snprint(range, sizeof(range), "Range: bytes=%lld-%llud", off, (off+len)-1);
	if((fd = exch(wdf->sess, params, nil, -1, -1)) < 0)
		return -1;

	p = buf;
	got = 0;
	while(len > 0 && (n = read(fd, p+got, len-got)) > 0)
		got += n;
	if(off+got > wdf->length){
		fprint(2, "wd_read: file has grown %lld -> %lld\n", wdf->length, off+got);
		wdf->length = off+got;
	}

	close(fd);
	return got;
}

/* partial put, appears to be fairly widely supported */
int
wd_pwrite(Wdfile *wdf, void *buf, long len, uvlong off)
{
	int fd;
	char range[128];
	Param params[] = {
		{ "url", 		wdf->path },
		{ "request",		"PUT" },
		{ "contenttype",	"application/octet-stream" },
		{ "headers",		range },
		{ nil, nil },
	};

	if(Wddebug){
		fprint(2, "wr_pwrite: path=%s len=%ld off=%lld\n", wdf->path, len, off);
		fprint(2, "	filecached=%s\n", (wdf->tmpfd)? "yes": "no");
	}
	wdf->modified = 1;
	if(off+len > wdf->length)
		wdf->length = off+len;

	if(wdf->tmpfd)
		return pwrite(wdf->tmpfd, buf, len, off);

	snprint(range, sizeof(range), "Content-Range: bytes %lld-%llud/%llud",
		off, (off+len)-1, off+len);
	if((fd = exch(wdf->sess, params, buf, len, -1)) < 0)
		return -1;
	close(fd);
	return len;
}

int
wd_close(Wdfile *wdf)
{
	int fd, rc;
	char range[128];
	Param params[] = {
		{ "url", 		wdf->path },
		{ "request",		"PUT" },
		{ "contenttype",	"application/octet-stream" },
		{ "headers",		range },
		{ nil, nil },
	};

	if(Wddebug){
		fprint(2, "wr_close: path=%s modified=%d length=%lld\n",
			wdf->path, wdf->modified, wdf->length);
		fprint(2, "	filecached=%s\n", (wdf->tmpfd)? "yes": "no");
	}

	rc = 0;
	if(wdf->modified){
		invalidate(wdf->sess, wdf->path);
		if(wdf->tmpfd){
			snprint(range, sizeof(range),
				"Content-Range: bytes 0-%llud/%llud",
				wdf->length-1, wdf->length);
			if(Wddebug)
				fprint(2, "wr_close: modified path=%s len=%lld\n",
					wdf->path, wdf->length);
			if((fd = exch(wdf->sess, params, nil, -1, wdf->tmpfd)) < 0)
				rc = -1;
			else
				close(fd);
		}
	}

	if(wdf->tmpfd)
		close(wdf->tmpfd);

	free(wdf->path);
	free(wdf->type);
	free(wdf);
	return rc;
}

int
decdir(Xml *xp, Elem *multi, Dir *d)
{
	char *p;
	Elem *resp, *ep;

	if((resp = decend(xp, multi, "response")) == nil){
		werrstr("wd_readdir: <response> not found");
		return -1;
	}

	if((p = pcdata(xp, resp, "href")) == nil){
		werrstr("wd_readdir: <href> not found");
		return -1;
	}
	d->name = estrdup9p(href2name(p));

	for(ep = resp; ep; ep = ep->next)
		if(status(ep) == 200 && ep->child != nil)
			break;

	if(xml_find(xp, ep->child, "prop/resourcetype/collection") != nil)
		d->mode = DMDIR | 0777;
	else
		d->mode = 0666;

	if((p = pcdata(xp, ep->child, "prop/executable")) != nil)
		if(testbool(p) == 1)
			d->mode |= 0111;

	if((p = pcdata(xp, ep->child, "prop/isreadonly")) != nil)
		if(testbool(p) == 1)
			d->mode &= ~0222;

	if((p = pcdata(xp, ep->child, "prop/creator-displayname")) != nil)
		d->uid = estrdup9p(p);
	else
		d->uid = estrdup9p("unknown");
	d->muid = estrdup9p(d->uid);

	if((p = pcdata(xp, ep->child, "prop/version-name")) != nil){
		if((d->gid = smprint("r%s", p)) == nil)
			sysfatal("No memory\n");
	}
	else
		d->gid = estrdup9p("unknown");

	if((d->mode & DMDIR) == 0 &&
	   (p = pcdata(xp, ep->child, "prop/getcontentlength")) != nil)
		d->length = strtoul(p, nil, 10);

	if((p = pcdata(xp, ep->child, "prop/getlastmodified")) != nil)
		d->mtime = ctime2sec(p);

	if((p = pcdata(xp, ep->child, "prop/lastaccessed")) != nil)
		d->atime = ctime2sec(p);

	d->qid = mkqid(d->name, d->mtime, d->mode & DMDIR);
	return 0;
}

/*
 * basicialy a special version of wd_dirscan which parses the
 * psudo root directory that subversion can generate if you
 * enable the 'svnParentPath' config option. This is xml_HTML
 * rather than webdab so we must use GET rather than PROPFIND
 * but the idea is similar.
 *
 * Its not clear how widespread the use of this is, but it
 * is handy for me (Steve) at my work.
 */
int
svn_rootscan(Session *s, char *path, int dirscan, Dir **dp)
{
	Dir *d;
	long now;
	Xml *reply;
	int num, fd;
	char *p, *name;
	Elem *ep, *index;
	Param params[] = {
		{ "url", 			path },
		{ "request",		"GET" },
		{ nil, nil },
	};

	now = time(nil);
	if(Wddebug)
		fprint(2, "svnroot: path=%s dirscan=%d\n", path, dirscan);

	if(dirscan == 0){
		*dp = emalloc9p(sizeof(Dir) * 1);
		d = *dp;
		d->name = estrdup9p(path);
		d->mode = 0755|DMDIR;
		d->uid = estrdup9p("svn");
		d->gid = estrdup9p("svn");
		d->muid = estrdup9p("svn");
		d->mtime = now;
		d->atime = now;
		d->qid = mkqid(d->name, d->mtime, d->mode & DMDIR);
		return 1;
	}

	if((fd = exch(s, params, nil, -1, -1)) < 0)
		return -1;

	if((reply = xml_parse(fd, 0, Fcrushwhite|Fstripnamespace)) == nil){
		werrstr("svn_rootscan: xml_ML parse failed - %r");
		close(fd);
		return -1;
	}

	if(Propdebug || Wddebug > 1)
		xml_print(reply, 0, 1);

	if((index = xml_find(reply, reply->root, "svn/index/dir")) == nil){
		werrstr("svn_rootscan: not an SVN 'svnParentPath' index - %r");
		xml_free(reply);
		close(fd);
		return -1;
	}

	num = 0;
	for(ep = index; ep; ep = ep->next)
		if(xml_value(ep, "href") != nil)
			num++;

	*dp = emalloc9p(sizeof(Dir) * num);
	d = *dp;

	for(ep = index; ep; ep = ep->next)
		if((name = xml_value(ep, "href")) != nil){
			d->name = estrdup9p(name);
			if((p = strrchr(d->name, '/')) != nil)
				*p = 0;
			d->mode = 0755|DMDIR;
			d->uid = estrdup9p("svn");
			d->gid = estrdup9p("svn");
			d->muid = estrdup9p("svn");
			d->mtime = now;
			d->atime = now;
			d->qid = mkqid(d->name, d->mtime, d->mode & DMDIR);
			d++;
		}

	xml_free(reply);
	close(fd);
	return d - *dp;
}
	
int
_wd_dirread(Session *s, char *path, int dirscan, Dir **dp)
{
	Dir *d;
	int fd, num;
	char *req, depth[128];
	Xml *reply;
	Elem *first, *multi;
	Param params[] = {
		{ "url", 		path },
		{ "request",		"PROPFIND" },
		{ "headers",		depth },
		{ "contenttype",	"application/xml" },
		{ nil, nil },
	};
	static char *getdir = 
		"<?xml version='1.0' encoding='utf-8'?> "
		"<propfind xmlns='DAV:' >"
			"<prop> "
				"<version-name /> "
				"<isreadonly /> "		/* Microsoft extension */
				"<executable /> "		/* Microsoft extension */
				"<lastaccessed /> "		/* Microsoft extension */
				"<creator-displayname /> "
				"<getcontentlength /> "
				"<getlastmodified /> "
				"<resourcetype /> "
			"</prop> "
		"</propfind>\r\n";
	
	static char *getall = 
		"<?xml version='1.0' encoding='utf-8' ?> "
		"<propfind xmlns='DAV:' "
		    "xmlns:in='http://sapportals.com/xmlns/cm/webdav'> "
			"<allprop/> "
			"<in:include> "
				"<isreadonly /> "		/* Microsoft extension */
				"<executable /> "		/* Microsoft extension */
				"<lastaccessed /> "		/* Microsoft extension */
			"</in:include> "
		"</propfind>\r\n";

	if(Propdebug)
		req = getall;
	else
		req = getdir;

	if(Wddebug)
		fprint(2, "wd_readdir: path=%s dirscan=%d\n", path, dirscan);

	if(dirscan)
		strcpy(depth, "Depth: 1");
	else
		strcpy(depth, "Depth: 0");

	if((fd = exch(s, params, req, -1, -1)) < 0)
		return -1;

	if((reply = xml_parse(fd, 0, Fcrushwhite|Fstripnamespace)) == nil){
		werrstr("wd_readdir: xml_ML parse failed - %r");
		close(fd);
		return -1;
	}

	if(Propdebug || Wddebug > 1)
		xml_print(reply, 0, 1);


	if(reply->root->child == nil){
		werrstr("wd_readdir: no reply data  '%s'  '%s'", depth, path);
		xml_free(reply);
		close(fd);
		if(strcmp(path, "/") == 0)
			if((num = svn_rootscan(s, path, dirscan, dp)) != -1)
				return num;
		return -1;
	}

	/*
	 * skip first response on directory scans
	 * as that is the info for the directory itself
	 * FIxml_ME: can we always assume the first one is to be skipped?
	 */
	if(dirscan)
		first = reply->root->child->next;
	else
		first = reply->root->child;

	num = 0;
	for(multi = first; multi; multi = multi->next)
		num++;

	*dp = emalloc9p(sizeof(Dir) * num);
	d = *dp;

	for(multi = first; multi; multi = multi->next){
		if(decdir(reply, multi, d) < 0){
			close(fd);
			/* let the caller free the dir */
			xml_free(reply);
			return -1;
		}

		if(Wddebug > 1){
			fprint(2, "    name=%s uid=%s gid=%s muid=%s len=%lld\n",
				d->name, d->uid, d->gid, d->muid, d->length);
			fprint(2, "	mode=%luo q.type=0x%ux q.vers=%lud q.path=0x%llux\n",
				d->mode, d->qid.type, d->qid.vers, d->qid.path);
			fprint(2, "	mtime=%s", ctime(d->mtime));
			fprint(2, "	atime=%s", ctime(d->atime));
		}
		d++;
	}
	close(fd);
	xml_free(reply);
	return d - *dp;

}

int
wd_mkcol(Session *s, char *path)
{
	int fd;
	Param params[] = {
		{ "url", 	path },
		{ "request",	"MKCOL" },
		{ nil, nil },
	};

	if((fd = exch(s, params, nil, -1, -1)) < 0)
		return -1;
	invalidate(s, path);
	close(fd);
	return 0;
}

int
wd_delete(Session *s, char *path)
{
	int fd;
	Param params[] = {
		{ "url", 	path },
		{ "request",	"DELETE" },
		{ nil, nil },
	};

	if(Wddebug)
		fprint(2, "wd_delete: path=%s\n", path);
	if((fd = exch(s, params, nil, -1, -1)) < 0)
		return -1;
	invalidate(s, path);
	close(fd);
	return 0;
}

int
wd_move(Session *s, char *old, char *new)
{
	int fd;
	char dest[128];
	Param params[] = {
		{ "url", 	old },
		{ "request",	"MOVE" },
		{ "headers",	dest },
		{ nil, nil },
	};

	snprint(dest, sizeof(dest), "Destination: %s%s", s->base, new);

	if((fd = exch(s, params, nil, -1, -1)) < 0)
		return -1;
	invalidate(s, old);
	invalidate(s, new);
	close(fd);
	return 0;
}

int
wd_wstat(Session *s, char *path, char *uid, vlong length,
	    long mtime, long atime, int mode)
{
	Fmt f;
	int fd;
	char tmp[32], *buf, *state;
	Param params[] = {
		{ "url", 		path },
		{ "request",		"PROPPATCH" },
		{ "contenttype",	"application/xml" },
		{ nil, nil },
	};
	static char *prelude =
		"<?xml version='1.0'?>\n"
		"<propertyupdate xmlns='DAV:'>\n"
		"	<set>\n"
		"		<prop>\n";
	static char *prologue =
		"		</prop>\n"
		"	</set>\n"
		"</propertyupdate>\n";


	fmtstrinit(&f);
	fmtprint(&f, prelude);
	if(uid && uid[0])	/* DeltaV only, and probably illegal */
		fmtprint(&f, "<creator-displayname>%s</creator-displayname>\n", uid);

	if(~length)		/* probably illegal - its a live property */
		fmtprint(&f, "<getcontentlength>%lld</getcontentlength/>\n", length);

	if(~mtime)		/* probably illegal - its a live property */
		fmtprint(&f, "<getlastmodified>%s</getlastmodified>\n",
			sec2rfc2068(mtime, tmp, sizeof(tmp)));

	if(~atime)		/* not widely supported */
		fmtprint(&f, "<lastaccess>%s</lastaccess>\n",
			sec2rfc2518(atime, tmp, sizeof(tmp)));

	if(~mode){		/* not widely supported */
// FIxml_ME: booleans, should they be T/F true/false or 1/0 ???
		state = ((mode & 0222) == 0)? "1": "0";
		fmtprint(&f, "<isreadonly>%s</isreadonly>\n", state);

		state = ((mode & 0111) != 0)? "1": "0";
		fmtprint(&f, "<executable>%s</executable>\n", state);
	}
	fmtprint(&f, prologue);
	buf = fmtstrflush(&f);

	if(Propdebug)
		fprint(2, "%s\n", buf);

	if((fd = exch(s, params, buf, strlen(buf), -1)) < 0){
		free(buf);
		return -1;
	}

	invalidate(s, path);
	free(buf);
	close(fd);
	return 0;
}
